@@ -4,17 +4,19 @@ import json
 import pickle
 import secrets
 import uuid
+import dash_core_components as dcc
 import dash
 import dash_html_components as html
 import dash.dependencies as dd
 import plotly
 
 from dash.dependencies import Input, State, Output, MATCH, ALL, ALLSMALLER, _Wildcard
-from dash.exceptions import PreventUpdate
+from dash.development.base_component import Component
 from flask import session
 from flask_caching.backends import FileSystemCache
-from more_itertools import unique_everseen, flatten
-from json.decoder import JSONDecodeError
+from more_itertools import flatten
+from collections import defaultdict
+from typing import Dict
 
 _wildcard_mappings = {ALL: "<ALL>", MATCH: "<MATCH>", ALLSMALLER: "<ALLSMALLER>"}
 _wildcard_values = list(_wildcard_mappings.values())
@@ -28,6 +30,7 @@ class DashProxy(dash.Dash):
     def __init__(self, *args, transforms=None, **kwargs):
         super().__init__(*args, **kwargs)
         self.callbacks = []
+        self.clientside_callbacks = []
         self.arg_types = [dd.Output, dd.Input, dd.State]
         self.transforms = transforms if transforms is not None else []
         # Do the transform initialization.
@@ -70,14 +73,21 @@ class DashProxy(dash.Dash):
 
         return wrapper
 
+    def clientside_callback(self, clientside_function, *args, **kwargs):
+        self.clientside_callbacks.append(dict(clientside_function=clientside_function, args=args, kwargs=kwargs))
+
     def _register_callbacks(self, app=None):
         callbacks = list(self._resolve_callbacks())
+        app = super() if app is None else app
         for callback in callbacks:
             outputs = callback[dd.Output][0] if len(callback[dd.Output]) == 1 else callback[dd.Output]
-            if app is None:
-                super().callback(outputs, callback[dd.Input], callback[dd.State], **callback["kwargs"])(callback["f"])
-            else:
-                app.callback(outputs, callback[dd.Input], callback[dd.State], **callback["kwargs"])(callback["f"])
+            app.callback(outputs, callback[dd.Input], callback[dd.State], **callback["kwargs"])(callback["f"])
+
+    def _register_clientside_callbacks(self, app=None):
+        clientside_callbacks = list(self._resolve_clientside_callbacks())
+        app = super() if app is None else app
+        for callback in clientside_callbacks:
+            app.clientside_callback(callback["clientside_function"], *callback["args"], **callback["kwargs"])
 
     def _layout_value(self):
         layout = self._layout() if self._layout_is_function else self._layout
@@ -91,6 +101,7 @@ class DashProxy(dash.Dash):
         """
         # Register the callbacks.
         self._register_callbacks()
+        self._register_clientside_callbacks()
         # Proceed as normally.
         super()._setup_server()
         # Set session secret. Used by some subclasses.
@@ -106,6 +117,30 @@ class DashProxy(dash.Dash):
             callbacks = transform.apply(callbacks)
         return callbacks
 
+    def _resolve_clientside_callbacks(self):
+        """
+         This method resolves the clientside_callbacks, i.e. it applies the clientside_callbacks injections.
+        """
+        clientside_callbacks = self.clientside_callbacks
+        for transform in self.transforms:
+            clientside_callbacks = transform.apply_clientside(clientside_callbacks)
+        return clientside_callbacks
+
+    def hijack(self, app: dash.Dash):
+        # Change properties.
+        app.config.update(self.config)
+        app.title = self.title
+        app.index_string = self.index_string
+        # Inject layout.
+        app.layout = html.Div()  # fool layout validator
+        app._layout_value = self._layout_value
+        # Register callbacks.
+        self._register_callbacks(app)
+        self._register_clientside_callbacks(app)
+        # Setup secret.
+        if not app.server.secret_key:
+            app.server.secret_key = secrets.token_urlsafe(16)
+
 
 def _get_session_id(session_key=None):
     session_key = "session_id" if session_key is None else session_key
@@ -154,6 +189,9 @@ class DashTransform:
     def apply(self, callbacks):
         raise NotImplementedError()
 
+    def apply_clientside(self, clientside_callbacks):
+        return clientside_callbacks
+
     def layout(self, layout, layout_is_function):
         return layout
 
@@ -253,107 +291,124 @@ def trigger_filter(args):
 
 # endregion
 
-# region Group transform
+# region Multiplexer transform
 
-class GroupTransform(DashTransform):
+def _mp_id(output: Output, idx: int) -> Dict[str, str]:
+    return dict(id=output.component_id, prop=output.component_property, idx=idx)
+
+
+def _mp_element(mp_id: Dict[str, str]) -> dcc.Store:
+    return dcc.Store(id=mp_id)
+
+
+def _mp_prop() -> str:
+    return "data"
+
+
+def inject_proxies(node, proxy_map):
+    if not hasattr(node, "children") or node.children is None:
+        return
+    children = _as_list(node.children)
+    modified = False
+    for i, child in enumerate(children):
+        # Attach the proxy components as children of the original component to ensure dcc.Loading works.
+        if not hasattr(child, "id"):
+            continue
+        for key in proxy_map:
+            if not child.id == key:
+                continue
+            children[i] = html.Div([child] + proxy_map[key])
+            modified = True
+    if modified:
+        node.children = children
+
+
+def inject_proxies_recursively(node, proxy_map):
+    if not hasattr(node, "children") or node.children is None:
+        return
+    children = _as_list(node.children)
+    modified = False
+    for i, child in enumerate(children):
+        # Do recursion.
+        inject_proxies(child, proxy_map)
+        # Attach the proxy components as children of the original component to ensure dcc.Loading works.
+        if not hasattr(child, "id"):
+            continue
+        for key in proxy_map:
+            if not child.id == key:
+                continue
+            children[i] = html.Div([child] + proxy_map[key])
+            modified = True
+    if modified:
+        node.children = children
+
+
+class MultiplexerTransform(DashTransform):
+    """
+    The MultiplexerTransform makes it possible to target an output by callbacks multiple times. Under the hood, proxy
+    components (dcc.Store) are used, and the proxy_location keyword argument determines where these proxies are placed.
+    The default value "inplace" means that the original component is replace by a Div element that wraps the original
+    component and its proxies. The means that dcc.Loading will work, but it also means that if you replace the layout
+    of a component higher in the tree in a callback, you might end up deleting the proxies (!), which will break your
+    app. For this particular case, you can set proxy_location to a custom component (or None to use the layout root),
+    to place the proxies here instead. To use dcc.Loading for this particular case, the proxy_location must be wrapped.
+    """
+
+    def __init__(self, proxy_location="inplace"):
+        self.initialized = False
+        self.proxy_location = proxy_location
+        self.proxy_map = defaultdict(lambda: [])
+        self.app = DashProxy()
+
+    def layout(self, layout, layout_is_function):
+        if layout_is_function or not self.initialized:
+            # Inject proxies in a user defined component.
+            if self.proxy_location == "inplace":
+                inject_proxies_recursively(layout, self.proxy_map)
+            # Inject proxies in a component, either user defined or top level.
+            else:
+                target = self.proxy_location if isinstance(self.proxy_location, Component) else layout
+                proxies = list(flatten(list(self.proxy_map.values())))
+                target.children = _as_list(target.children) + proxies
+            self.initialized = True
+        return layout
 
     def apply(self, callbacks):
-        groups = {}
-        # Figure out which callbacks to group together.
-        grouped_callbacks = []
-        for i in range(len(callbacks)):
-            key = callbacks[i]["kwargs"].pop("group", None)
-            if key:
-                if key not in groups:
-                    groups[key] = []
-                groups[key].append(i)
-            else:
-                grouped_callbacks.append(callbacks[i])
-        # Do the grouping.
-        for key in groups:
-            grouped_callback = _combine_callbacks([callbacks[i] for i in groups[key]])
-            grouped_callbacks.append(grouped_callback)
-
-        return grouped_callbacks
-
-
-# NOTE: No performance considerations what so ever. Just an initial proof-of-concept implementation.
-def _combine_callbacks(callbacks):
-    inputs, input_props, input_prop_lists, input_mappings = _prep_props(callbacks, dd.Input)
-    states, state_props, state_prop_lists, state_mappings = _prep_props(callbacks, dd.State)
-    outputs, output_props, output_prop_lists, output_mappings = _prep_props(callbacks, dd.Output)
-    # TODO: What kwargs to use?
-    kwargs = callbacks[0]["kwargs"]
-    multi_output = any([callback["multi_output"] for callback in callbacks])
-    if not multi_output:
-        all_outputs = []
+        # Group by output.
+        output_map = defaultdict(list)
         for callback in callbacks:
-            all_outputs += callback[dd.Output]
-        multi_output = len(list(set(all_outputs))) > 1
-
-    # TODO: There might be a scope issue here
-    def wrapper(*args):
-        local_inputs = list(args)[:len(inputs)]
-        local_states = list(args)[len(inputs):]
-        if len(dash.callback_context.triggered) == 0:
-            raise PreventUpdate
-        prop_id = dash.callback_context.triggered[0]['prop_id']
-        output_values = [dash.no_update] * len(outputs)
-        for i, entry in enumerate(input_prop_lists):
-            # Check if the trigger is an input of the callback.
-            match = False
-            for item in entry:
-                # Check for exact matches.
-                match = item == prop_id
-                if match:
-                    break
-                # Check for wild card matches.
-                if any([wildcard_value in item for wildcard_value in _wildcard_values]):
-                    try:
-                        props = json.loads(prop_id.split(".")[0])
-                        item_props = json.loads(item.split(".")[0])
-                        prop_match = True
-                        for key in props:
-                            if item_props[key] not in _wildcard_values:
-                                prop_match = prop_match and item_props[key] == props[key]
-                            # TODO: Make checks here, no checks (as now) is only valid for ALL
-                        if prop_match:
-                            match = True
-                            break
-                    except JSONDecodeError:
-                        continue
-            if not match:
+            for output in callback[Output]:
+                output_map[output].append(callback)
+        # Apply multiplexer where needed.
+        for output in output_map:
+            # If there is only one output, multiplexing is not needed.
+            if len(output_map[output]) == 1:
                 continue
-            # Trigger the callback function.
-            try:
-                inputs_i = [local_inputs[j] for j in input_mappings[i]]
-                states_i = [local_states[j] for j in state_mappings[i]]
-                outputs_i = callbacks[i]["f"](*inputs_i, *states_i)
-                if not callbacks[i]["multi_output"]:  # len(callbacks[i][Output]) == 1:  TODO: Is this right?
-                    outputs_i = [outputs_i]
-                for j, item in enumerate(outputs_i):
-                    output_values[output_mappings[i][j]] = outputs_i[j]
-            except PreventUpdate:
-                continue
-        # Check if an update is needed.
-        if all([item == dash.no_update for item in output_values]):
-            raise PreventUpdate
-        # Return the combined output.
-        return output_values if multi_output else output_values[0]  # TODO: Check for multi output here?
+            self._apply_multiplexer(output, output_map[output])
 
-    return {dd.Output: outputs, dd.Input: inputs, "sorted_args": outputs + inputs + states,
-            "f": wrapper, dd.State: states, "kwargs": kwargs, "multi_output": multi_output}
+        return callbacks
 
+    def apply_clientside(self, clientside_callbacks):
+        return clientside_callbacks + self.app.clientside_callbacks
 
-def _prep_props(callbacks, key):
-    all = []
-    for callback in callbacks:
-        all.extend(callback[key])
-    all = list(unique_everseen(all))
-    props = [_create_callback_id(item) for item in all]
-    prop_lists = [[_create_callback_id(item) for item in callback[key]] for callback in callbacks]
-    mappings = [[props.index(item) for item in l] for l in prop_lists]
-    return all, props, prop_lists, mappings
+    def _apply_multiplexer(self, output, callbacks):
+        inputs = []
+        proxies = []
+        for i, callback in enumerate(callbacks):
+            # Create proxy element.
+            proxies.append(_mp_element(_mp_id(output, i)))
+            # Assign proxy element as output.
+            callback[Output][callback[Output].index(output)] = Output(proxies[-1].id, _mp_prop())
+            # Create proxy input.
+            inputs.append(Input(_mp_id(output, ALL), _mp_prop()))
+        # Collect proxy elements to add to layout.
+        self.proxy_map[output.component_id].extend(proxies)
+        # Create multiplexer callback. Clientside for best performance. TODO: Is this robust?
+        self.app.clientside_callback("""
+            function(){
+                return dash_clientside.callback_context.triggered[0].value;
+            }
+        """, output, inputs, prevent_initial_call=True)
 
 
 # endregion
@@ -581,7 +636,7 @@ class NoOutputTransform(DashTransform):
 class Dash(DashProxy):
     def __init__(self, *args, output_defaults=None, **kwargs):
         output_defaults = dict(backend=None, session_check=True) if output_defaults is None else output_defaults
-        transforms = [TriggerTransform(), NoOutputTransform(), GroupTransform(),
+        transforms = [TriggerTransform(), MultiplexerTransform(), NoOutputTransform(),
                       ServersideOutputTransform(**output_defaults)]
         super().__init__(*args, transforms=transforms, **kwargs)
 
